/** Part of the LightZPng library package, v1.0
 ** 2005 Jere Sanisalo & Housemarque Inc
 **
 ** License for this file (MIT):
 **
 ** Copyright (c) 2005 Jere Sanisalo & Housemarque Inc
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a
 ** copy of this software and associated documentation files (the "Software"),
 ** to deal in the Software without restriction, including without limitation
 ** the rights to use, copy, modify, merge, publish, distribute, sublicense,
 ** and/or sell copies of the Software, and to permit persons to whom the
 ** Software is furnished to do so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in
 ** all copies or substantial portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 ** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 ** LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 ** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 **/
#include <stdio.h>
#include <string.h>
#include "LightPng.h"
#include "LightZ.h"

// Reads a file fully (return null on error)
static char *ReadFile( const char *fn, int &dest_size )
{
	FILE *f = fopen( fn, "rb" );
	if ( f == 0 )
		return 0;

	fseek( f, 0, SEEK_END );
	dest_size = ftell( f );
	fseek( f, 0, SEEK_SET );

	char *data = new char[ dest_size + 1 ];
	if ( fread( data, 1, dest_size, f ) != dest_size )
	{
		delete [] data;
		data = 0;
	}
	fclose( f );

	return data;
}

// Reads and unpacks a zlib "compress" compressed file
static void TestZFile( const char *fn, const char *fn_out )
{
	printf( "\nTesting Z-file: '%s'\n", fn );
	
	int source_len;
	char *source = ReadFile( fn, source_len );
	if ( source == 0 )
	{
		printf( "Error reading file!" );
		return;
	}

	unsigned char *dest = 0;
	int dest_len = 0;

	// Test the uncompression
	const char *err_msg = LZ_Inflate( source, source_len, &dest, &dest_len );
	if ( err_msg != 0 )
		printf( "ERROR: %s\n", err_msg );
	else
		printf( "Ok!\n" );
	printf( "Uncompressed size: %d\n", dest_len );

	// Write out the uncompressed file
	FILE *f = fopen( fn_out, "wb" );
	if ( f != 0 )
	{
		if ( dest != 0 && dest_len > 0 )
			fwrite( dest, 1, dest_len, f );
		fclose( f );
	}

	// Release the memory
	delete [] source;
	if ( dest != 0 )
		delete [] dest;
}

// Writes a byte to the output
static inline void WriteByte( FILE *f, unsigned char b )
{
	fwrite( &b, 1, sizeof( b ), f );
}

// Writes an image as TGA (32bit)
static void WriteTga( const char *fn, LPNG_Image *img )
{
	if ( img == 0 )
		return;

	printf( "Writing '%s'..\n", fn );
	if ( img->has_palette && img->palette == 0 )
	{
		printf( "Unable to write TGA: image has palatte, but palette data is missing\n" );
		return;
	}
	if ( img->data == 0 )
	{
		printf( "Unable to write TGA: missing image data\n" );
		return;
	}

	// Open the file
	FILE *f = fopen( fn, "wb" );
	if ( f == 0 )
	{
		printf( "Unable to write TGA: error opening the file for writing\n" );
		return;
	}

	int w = img->width;
	int h = img->height;

	// Write the header
	WriteByte( f, 0 ); // identsize
	WriteByte( f, 0 ); // colormaptype
	WriteByte( f, 2 ); // imagetype

	// Color map specs
	WriteByte( f, 0 );
	WriteByte( f, 0 );
	WriteByte( f, 0 );
	WriteByte( f, 0 );
	WriteByte( f, 0 );

	WriteByte( f, 0 ); // X-origin
	WriteByte( f, 0 ); // X-origin
	WriteByte( f, ((h-1) >> 0) & 0xff ); // Y-origin
	WriteByte( f, ((h-1) >> 8) & 0xff ); // Y-origin
	WriteByte( f, (w >> 0) & 0xff ); // Width
	WriteByte( f, (w >> 8) & 0xff ); // Width
	WriteByte( f, (h >> 0) & 0xff ); // Height
	WriteByte( f, (h >> 8) & 0xff ); // Height
	WriteByte( f, 32 ); // Pixel depth
	WriteByte( f, 8 | (0<<4) | (1<<5) ); // Image desc

	// Write the image data
	LPNG_Image::Color *palette = img->palette;
	unsigned char *data = img->data;

	for ( int y = 0; y < h; ++y )
	{
		for ( int x = 0; x < w; ++x )
		{
			unsigned char a, r, g, b;

			if ( img->has_palette )
			{
				LPNG_Image::Color col = palette[ *data++ ];
				a = (col >> 24) & 0xff;
				r = (col >> 16) & 0xff;
				g = (col >> 8) & 0xff;
				b = (col >> 0) & 0xff;
			}
			else
			{
				a = *data++;
				r = *data++;
				g = *data++;
				b = *data++;
			}

			WriteByte( f, b );
			WriteByte( f, g );
			WriteByte( f, r );
			WriteByte( f, a );
		}
	}

	fclose( f );
}

// Reads and unpacks the ginve .png file
static void TestPngFile( const char *fn, const char *tga_fn )
{
	printf( "\nTesting PNG-file: '%s'\n", fn );
	
	int source_len;
	char *source = ReadFile( fn, source_len );
	if ( source == 0 )
	{
		printf( "Error reading file!" );
		return;
	}

	// Open it
	LPNG_Image *img = LPNG_Create( source, source_len );
	if ( img == 0 )
	{
		printf( "Error opening the image!\n" );
		return;
	}

	printf( "Image opened ok\n" );
	printf( "Width : %d\n", img->width );
	printf( "Height: %d\n", img->height );
	printf( "Has palette: %s\n", img->has_palette ? "yes" : "no" );

	WriteTga( tga_fn, img );

	delete img;
}

int main( int argc, char *argv[] )
{
	TestZFile( "test1.z", "test1.uz" );
	TestZFile( "test2.z", "test2.uz" );
	TestZFile( "test3.z", "test3.uz" );
	TestZFile( "test4.z", "test4.uz" );
	TestZFile( "test5.z", "test5.uz" );
	TestPngFile( "test_24bit.png", "test_24bit.tga" );
	TestPngFile( "test_32bit.png", "test_32bit.tga" );
	TestPngFile( "test_8bit_alpha.png", "test_8bit_alpha.tga" );
	return 0;
}
